from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, text, Index, UniqueConstraint, DateTime, BigInteger
from sqlalchemy import event
# from sqlalchemy import TIMESTAMP, JSON

Base = declarative_base()
metadata = Base.metadata


class ScheduleOrm(Base):
    '''
    schedule和schedule strategy从职责的角度最好分开表，但是从触发更新时间和扫描模式更适合合并，所以先合并一个表，读到内存后再以对象存在
    '''
    __tablename__ = 'schedule'
    __table_args__ = (
        UniqueConstraint('key', name='uniq_key'),
        Index('idx_updated_time', 'updated_time'),
        {
            'mysql_collate': 'utf8_general_ci',
            'mysql_engine': 'InnoDB',
            'mysql_default_charset': 'utf8',
        },
    )

    id = Column(BigInteger, primary_key=True, nullable=False, autoincrement=True)
    key = Column(String(100, collation='utf8_general_ci'), unique=True, nullable=False)
    name = Column(String(255, collation='utf8_general_ci'), nullable=False)
    sharding_key = Column(String(10, collation='utf8_general_ci'), nullable=False)
    region = Column(String(10, collation='utf8_general_ci'), nullable=False)
    stop_time = Column(BigInteger, nullable=False)
    version = Column(Integer, nullable=False)
    created_time = Column(DateTime(timezone=True), nullable=False, server_default=text('CURRENT_TIMESTAMP'))
    updated_time = Column(DateTime(timezone=True), nullable=False, server_default=text('CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP'))


@event.listens_for(ScheduleOrm, 'before_insert')
def do_before_insert(mapper, connect, target):
    print('do_before_insert')
    print(target.sharding_key)


@event.listens_for(ScheduleOrm, 'before_update')
def do_before_update(mapper, connection, target):
    print('do_before_update')
    print(target.sharding_key)


@event.listens_for(ScheduleOrm, 'load')
def receive_load(target, context):
    print('receive_load')
    print(target.sharding_key)


@event.listens_for(ScheduleOrm.sharding_key, 'modified')
def receive_modified(target, initiator):
    print('receive_bulk_replace')
    print(target.sharding_key)
    print(initiator)



import os

# DB
DB_TYPE = 'mysql'
DB_ROOT_PASSWORD = 'root'
DB_NAME = 'test_db'
DB_USER = 'ttt'
DB_PASSWORD = os.getenv('schedule_mysql_password_test', None)
DB_HOST_PORT = 3306
DB_HOST = os.getenv('schedule_mysql_host_test', None)

DB_CONFIG = {
    'DB_ENGINE_CONFIG_PREFIX': 'test.',
    'DB_ENGINE_CONFIG': {
        'test.url': f'mysql+pymysql://{DB_USER}:{DB_PASSWORD}@{DB_HOST}:{DB_HOST_PORT}/{DB_NAME}?charset=utf8',
        'test.isolation_level': 'REPEATABLE_READ',
        'test.max_overflow': 10,
        'test.pool_size': 5,
        'test.echo': False,
    },
    'DB_ENGINE_POOL_SIZE': 5,
    'DB_ENGINE_MAX_OVERFLOW': 10,
    'DB_SESSION_AUTOFLUSH': True,
    'DB_SESSION_AUTOCOMMIT': False,
}



import time
import uuid
import traceback
import threading
from queue import Queue
from ..common import tools
from loguru import logger

DEFAULT_KPI = {
    'cpu_count': 4,
    'cpu_count_logical': 8,
    'cpu_percent': 0.0,
    'mem_total': 8,
    'mem_used': 5,
    'mem_free': 2,
    'mem_percent': 0.0,
    'disk_percent': 0.0,
}

lock = threading.Lock()


class ScheduleCenter:
    """
    ScheduleCenter必须是单例
    """
    _slave_keys_limit = 10000  # 限制一个slave最多托管多少个key
    _slave_queue_bindings = {}  # {address1: Queue1, address2: Queue2, ...}
    _key_slave_mappings = {}  # {key1: address2, key2: address3, ...}
    _slave_keys_mappings = {}  # {address1: {key1, key2, ...}, address2: {key11, key12, ...}, ...}
    """
    example:
    kpi dict = {
        'cpu_count': 4,
        'cpu_count_logical': 8,
        'cpu_percent': 0.2,
        'mem_total': 8,
        'mem_used': 5,
        'mem_free': 2,
        'mem_percent': 0.9,
        'disk_percent': 0.2,
    }
    """
    _slave_kpi_mappings = {}  # {address1: {[kpi dict1]}, address2: {[kpi dict2]}}
    _secret_key_store = {}  # {address1: secret_key2, address2: secret_key3, ...}
    _checksum_store = {}  # {address1: checksum1, address2: checksum2, ...}
    _slave_heartbeats = {}  # {address1: heartbeat_ts1, address2: heartbeat_ts2, ...}
    _reallocate_keys = set()  # {key1, key2, ..., keyn}

    _instance = None

    def __new__(cls, *args, **kwargs):
        # If no instance of class already exits
        if cls._instance is None:
            cls._instance = object.__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self, *args, **kwargs):
        if self._initialized:
            return
        # set the attribute to `True` to not initialize again
        self._initialized = True

    def set_slave_keys_limit(self, slave_keys_limit):
        self._slave_keys_limit = slave_keys_limit

    def slave_heartbeats(self):
        # 仅供Master使用
        return self._slave_heartbeats

    def reallocate_keys(self):
        # 仅供Master使用
        return self._reallocate_keys

    def secret_key_store(self):
        # 仅供Master使用
        return self._secret_key_store

    def checksum_store(self):
        # 仅供Master使用
        return self._checksum_store

    def add_reallocate_keys(self, key_set):
        # 仅供Master使用
        self._reallocate_keys |= key_set

    def bind(self, address, reported_kpi=None, now=None):
        # 仅供Worker在初始化时使用
        """
        绑定，各个Worker初始化时调用ScheduleCenter的这个方法，用于告诉ScheduleCenter，自己可以接任务了；
        同时ScheduleCenter记录Worker和队列的绑定关系
        """
        if now is None:
            now = int(time.time())
        if reported_kpi is None:
            reported_kpi = DEFAULT_KPI
        try:
            lock.acquire()
            if address in self._slave_queue_bindings:
                return 100, f'address: {address} already in bindings'
            secret_key = uuid.uuid4().hex
            # 保存secret key
            self._secret_key_store[address] = secret_key
            # 对该address生成新Queue
            self._slave_queue_bindings[address] = Queue()
            # 更新心跳
            self._slave_heartbeats[address] = now
            # 默认kpi百分比指标为0
            self._slave_kpi_mappings[address] = reported_kpi
            # 初始化空列表
            self._slave_keys_mappings[address] = set()
            # 初始化checksum为0
            self._checksum_store[address] = 0
            return 200, secret_key
        finally:
            lock.release()

    def _route_address(self):
        lowest_avg = 100.0
        most_free_address = None
        for _address, _kpi in self._slave_kpi_mappings.items():
            cpu_percent = _kpi['cpu_percent']
            mem_percent = _kpi['mem_percent']
            disk_percent = _kpi['disk_percent']
            avg_percent = (cpu_percent + mem_percent + disk_percent) / 3
            if avg_percent <= lowest_avg:
                slave_key_set_num = len(self._slave_keys_mappings[_address])
                if slave_key_set_num < self._slave_keys_limit:
                    lowest_avg = avg_percent
                    most_free_address = _address
        return most_free_address

    def allocate(self, key, address=None):
        # 仅供Master使用，用于给slave分配schedule的key
        try:
            lock.acquire()
            if address is None:
                # 先检查在不在已分配的slave中
                address = self._key_slave_mappings.get(key, None)
            if address is None:
                address = self._route_address()
            if address is None:
                logger.error(f'no enough slave workers!!!')
                return False
            self._key_slave_mappings[key] = address
            slave_key_set = self._slave_keys_mappings[address]
            slave_key_set.add(key)
            last_checksum = self._checksum_store[address]
            new_checksum = last_checksum + int.from_bytes(str.encode(key), byteorder='little')
            self._checksum_store[address] = new_checksum
            # 把key推到对应的queue
            queue = self._slave_queue_bindings[address]
            queue.put(key)
            logger.debug(f'put key: {key}')
            return True
        finally:
            lock.release()

    def take(self, address, secret_key, checksum, reported_kpi, batch_num=1000, now=None):
        # 仅供Worker使用
        if now is None:
            now = int(time.time())
        """
        接消息（任务），Worker调用该方法接收消息，消息类型包含任务和心跳
        """
        try:
            store_key = self._secret_key_store.get(address, None)
            if store_key is None:
                return 403, f'your address: {address} not in store'
            if store_key != secret_key:
                return 401, f'your address: {address} secret_key: {secret_key} not match'
            store_checksum = self._checksum_store.get(address, None)
            if store_checksum is None:
                return 403, f'your address: {address} not in checksum store'
            code = 200
            if store_checksum != checksum:
                code = 300
                # return 401, f'your address: {address} checksum: {checksum} not match'
                logger.warning(f'your address: {address} checksum: {checksum} not match store_checksum: {store_checksum}')

            # 更新心跳
            self._slave_heartbeats[address] = now
            # 更新KPI
            self._slave_kpi_mappings[address] = reported_kpi
            address_queue = self._slave_queue_bindings[address]
            msg_list = tools.getn_from_queue(address_queue, batch_num)
            return code, msg_list
        except Exception as e:
            logger.error(f'{e}, {traceback.print_exc()}')
            return 404, f'exception: {e}'

    def unbind(self, address, secret_key):
        # Worker退出时或者Master检测超时后调用
        """
        解除绑定，Worker或定时检测程序调用该方法解除绑定，Master则清除绑定关系，并把原本该Worker的任务分发给其他Worker
        """
        try:
            lock.acquire()
            store_key = self._secret_key_store.get(address, None)
            if store_key is None:
                return 403, f'your address: {address} not in store'
            if store_key != secret_key:
                return 401, f'your address: {address} secret_key: {secret_key} not match'
            # 遍历key所在slave的address
            for _key, _address in self._key_slave_mappings.copy().items():
                if _address == address:
                    # time1 = time.time()
                    self._reallocate_keys.add(_key)
                    # time2 = time.time()
                    # print(f'_reallocate_keys add use time = {time2 - time1}')
                    del self._key_slave_mappings[_key]
                    # time3 = time.time()
                    # print(f'_key_slave_mappings del use time = {time3 - time2}')
            # 删除slave对应queue的绑定关系
            del self._slave_queue_bindings[address]
            # 删除slave的对应secret key的存储
            del self._secret_key_store[address]
            # 删除slave的对应checksum的存储
            del self._checksum_store[address]
            # 删除slave包含keys列表的映射关系
            del self._slave_keys_mappings[address]
            del self._slave_kpi_mappings[address]
            del self._slave_heartbeats[address]
        except Exception as e:
            logger.critical(f'{e}, {traceback.print_exc()}')
        finally:
            lock.release()



from flask import Flask
from flask import jsonify
from ..common.schedule_center import ScheduleCenter

app = Flask(__name__)

schedule_center = ScheduleCenter()


@app.route('/')
def index():
    response = {
        'slave_queue_bindings': {k: v.qsize() for k, v in schedule_center._slave_queue_bindings.items()},
        'key_slave_mappings': len(schedule_center._key_slave_mappings),
        'slave_keys_mappings': {k: len(v) for k, v in schedule_center._slave_keys_mappings.items()},
        'slave_kpi_mappings': schedule_center._slave_kpi_mappings,
        'secret_key_store': schedule_center._secret_key_store,
        'checksum_store': schedule_center._checksum_store,
        'slave_heartbeats': schedule_center._slave_heartbeats,
        'reallocate_keys': len(schedule_center._reallocate_keys),
    }
    return jsonify(response)


@app.route('/schedule/<schedule_key>')
def get(schedule_key):
    response = {
        'hello': 'world'
    }
    return jsonify(response)



import re
import time
import psutil
from queue import Empty


def getn_from_queue(queue, n, sleep_seconds=0.0001):
    """
    通用的在queue里面批量取n个值再返回，方便后续的批量操作
    """
    results = []
    try:
        while len(results) < n:
            results.append(queue.get(block=False))
    except Empty:
        time.sleep(sleep_seconds)
    return results


def get_kpi():
    # 查看cpu物理个数的信息
    cpu_count = psutil.cpu_count(logical=False)
    # CPU的使用率
    cpu_percent = psutil.cpu_percent()
    # 查看cpu逻辑个数的信息
    cpu_count_logical = psutil.cpu_count(logical=True)
    # 内存
    mem = psutil.virtual_memory()
    # 系统总计内存
    mem_total = float(mem.total) / 1024 / 1024 / 1024
    # 系统已经使用内存
    mem_used = float(mem.used) / 1024 / 1024 / 1024
    # 系统空闲内存
    mem_free = float(mem.free) / 1024 / 1024 / 1024
    mem_percent = mem[2]
    # 磁盘使用率
    disk_percent = psutil.disk_usage('/').percent
    return {
        'cpu_count': cpu_count,
        'cpu_count_logical': cpu_count_logical,
        'cpu_percent': cpu_percent,
        'mem_total': round(mem_total, 2),
        'mem_used': round(mem_used, 2),
        'mem_free': round(mem_free, 2),
        'mem_percent': mem_percent,
        'disk_percent': disk_percent,
    }


def validate_iso8601(str_val):
    """
    校验是不是iso8601的字符串时间
    """
    regex = r'^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$'
    match_iso8601 = re.compile(regex).match
    try:
        if match_iso8601(str_val) is not None:
            return True
    except Exception:
        pass
    return False


def validate_utc_iso8601(str_val):
    """
    校验是不是iso8601的utc字符串时间
    """
    is_valid_iso8601 = validate_iso8601(str_val)
    is_utc = str_val[-5:] == '00:00'
    if is_valid_iso8601 and is_utc:
        return True
    return False



import traceback
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from .db import BaseDbManager
from .schedule_orm import Base
from .schedule_orm import ScheduleOrm
from ...common.tools import validate_utc_iso8601
from loguru import logger


class ScheduleDbManager(BaseDbManager):

    def __init__(self, db_config: dict):
        super().__init__(Base, db_config)

    def make_repository(self, db_type, sharding_key):
        if db_type == 'mysql':
            return ScheduleRepository(self._session, sharding_key)
        raise Exception('unknown db_type: {db_type}')


class ScheduleRepositoryFactory:

    @staticmethod
    def make(db_config, db_type, sharding_key):
        return ScheduleDbManager(db_config).make_repository(db_type, sharding_key)


class ScheduleRepository:

    # _offset可以是更新时间的datetime，也可以是id

    _pool: None
    _sharding_key: str = None

    def __init__(self, pool, sharding_key):
        self._pool = pool
        self._sharding_key = sharding_key

    _schedule_base_entities = (
        ScheduleOrm.id,
        ScheduleOrm.key,
        ScheduleOrm.name,
        ScheduleOrm.sharding_key,
        ScheduleOrm.region,
        ScheduleOrm.stop_time,
        ScheduleOrm.version,
        ScheduleOrm.created_time,
        ScheduleOrm.updated_time,
    )

    def get_schedule_by_key(self, key):
        try:
            schedule = self._pool.query(ScheduleOrm)\
                .with_entities(*self._schedule_base_entities)\
                .filter(ScheduleOrm.schedule_key == key)\
                .first()
            self._pool.commit()
            return schedule
        except Exception as e:
            self._pool.rollback()
            logger.error(e)
        finally:
            pass

    def load_schedules_by_key_list(self, key_list):
        try:
            schedules = self._pool.query(ScheduleOrm)\
                .with_entities(*self._schedule_base_entities)\
                .filter(ScheduleOrm.schedule_key.in_(key_list))\
                .all()
            self._pool.commit()
            return schedules
        except Exception as e:
            self._pool.rollback()
            logger.error(e)
        finally:
            pass

    def load_schedules_incr(self, sort_field_name, last_offset, limit=11111):
        # 增量拉数据
        sort_field = getattr(ScheduleOrm, sort_field_name)
        all_schedules = []
        next_offset = last_offset
        try:
            corner_schedules = self._pool.query(ScheduleOrm)\
                .with_entities(*self._schedule_base_entities)\
                .filter(sort_field == last_offset)\
                .all()
            corner_schedules_len = len(corner_schedules)
            if corner_schedules_len >= limit:
                all_schedules = corner_schedules
                next_offset = next_offset + timedelta(seconds=1)
            else:
                # 如果schedule state num < limit，则拉取大于last_offset的schedule state，limit （schedule state num - limit）个，返回new schedule states
                upper_schedules = self._pool.query(ScheduleOrm)\
                    .with_entities(*self._schedule_base_entities)\
                    .filter(sort_field > last_offset)\
                    .order_by(sort_field.asc())\
                    .limit(limit - corner_schedules_len)\
                    .all()
                if len(upper_schedules) > 0:
                    next_offset = getattr(upper_schedules[-1], sort_field_name)
                else:
                    if corner_schedules_len > 0:
                        next_offset = next_offset + timedelta(seconds=1)
                all_schedules = corner_schedules + upper_schedules
        except Exception as e:
            self._pool.rollback()
            logger.error(f'{e}, {traceback.print_exc()}')
        finally:
            return all_schedules, next_offset

    # def load_limit_need_init_schedules(self, sort_field_name, limit=222):
    #     sort_field = getattr(ScheduleOrm, sort_field_name)
    #     try:
    #         need_init_schedules = self._pool.query(ScheduleOrm)\
    #             .with_entities(*self._schedule_base_entities)\
    #             .filter(sort_field == '')\
    #             .limit(limit)\
    #             .all()
    #     except Exception as e:
    #         self._pool.rollback()
    #         logger.error(f'{e}, {traceback.print_exc()}')
    #     finally:
    #         return need_init_schedules

    # def load_limit_need_run_schedules(self, sort_field_name, current_offset, last_offset, limit=11111):
    #     """
    #     这里的limit只能限制大概的范围，不能保证实际返回的量一定小于等于limit，因为有边缘情况需要加上，所以实际返回量有可能大于limit
    #     当last_offset为''时，转换为'2017-04-10T02:00:00+00:00'
    #     """
    #     if last_offset == '':
    #         last_offset = '2017-04-10T02:00:00+00:00'
    #     assert validate_utc_iso8601(last_offset), last_offset
    #     assert validate_utc_iso8601(current_offset), current_offset
    #     sort_field = getattr(ScheduleOrm, sort_field_name)
    #     all_schedules = []
    #     next_offset = last_offset
    #     try:
    #         # 先拉取等于last_offset的schedule states，并记录数量schedule state num
    #         corner_schedules = self._pool.query(ScheduleOrm)\
    #             .with_entities(*self._schedule_base_entities)\
    #             .filter(sort_field == last_offset)\
    #             .all()
    #         corner_schedules_len = len(corner_schedules)
    #         if corner_schedules_len >= limit:
    #             all_schedules = corner_schedules
    #             next_offset = (datetime.fromisoformat(last_offset) + timedelta(seconds=1)).astimezone(timezone.utc).isoformat()
    #         else:
    #             # 如果schedule state num < limit，则拉取大于last_offset的schedule state，limit （schedule state num - limit）个，返回new schedule states
    #             upper_schedules = self._pool.query(ScheduleOrm)\
    #                 .with_entities(*self._schedule_base_entities)\
    #                 .filter(sort_field > last_offset)\
    #                 .filter(sort_field <= current_offset)\
    #                 .order_by(sort_field.asc())\
    #                 .limit(limit - corner_schedules_len)\
    #                 .all()
    #             if len(upper_schedules) == 0:
    #                 all_schedules = corner_schedules
    #                 next_offset = current_offset
    #             else:
    #                 new_offset = getattr(upper_schedules[-1], sort_field_name)
    #                 all_schedules = corner_schedules + upper_schedules
    #                 next_offset = datetime.fromisoformat(new_offset).astimezone(timezone.utc).isoformat()
    #         self._pool.commit()
    #     except Exception as e:
    #         self._pool.rollback()
    #         logger.error(f'{e}, {traceback.print_exc()}')
    #     finally:
    #         return all_schedules, next_offset

    def close(self):
        self._pool.close()



import sys
import time
import traceback
import threading
from queue import Queue
from .register_manager import RegisterManager
from .slave_client import SlaveClient
from loguru import logger


class ShouldReconnectException(Exception):

    pass


class SlaveService(threading.Thread):

    def __init__(self, host, port, authkey, receive_queue, batch_num=3000):
        assert type(receive_queue) is Queue
        threading.Thread.__init__(self)
        manager = RegisterManager(host, port, authkey)
        self._client = SlaveClient(manager)
        self._receive_queue = receive_queue
        self._checksum = 0
        self._batch_num = batch_num
        self._is_active = True
        self._heartbeat_time = 99999999999
        self._heartbeat_timeout = 3
        self.daemon = True

    # def __init__(self, client, receive_queue, batch_num=1000):
    #     assert type(receive_queue) is Queue
    #     self._client = client
    #     self._receive_queue = receive_queue
    #     self._checksum = 0
    #     self._batch_num = batch_num
    #     self._is_active = True
    #     self._heartbeat_time = 99999999999
    #     self._heartbeat_timeout = 3

    def is_connecting(self, now=None):
        if now is None:
            now = int(time.time())
        if now - self._heartbeat_time > self._heartbeat_timeout:
            return False
        return True

    def run_once(self):
        code, result = self._client.take(self._checksum, self._batch_num)
        if code not in [200, 300]:
            print(result)
            if code == 403:
                raise ShouldReconnectException('should reconnect!')
            return False
        if code == 300:
            logger.warning(f'checksum not match {self._checksum}!!!!')
        msg_list = result
        for msg in msg_list:
            self._receive_queue.put(msg)
            self._checksum += int.from_bytes(str.encode(msg), byteorder='little')
            print(f'my _checksum = {self._checksum}, len = {self._receive_queue.qsize()}')
        return len(msg_list)

    def run(self):
        assert self._is_active is True
        self._client.bind()
        while self._is_active:
            try:
                msg_len = self.run_once()
                self._heartbeat_time = int(time.time())
                if msg_len == 0:
                    time.sleep(2)
            except EOFError as e:
                logger.error(f'EOFError when request_with_unique_key: {e}')
                self.reconnect()
            except BrokenPipeError as e:
                # [Errno 32] Broken pipe
                logger.error(f'BrokenPipeError when request_with_unique_key: {e}')
                self.reconnect()
            except ConnectionResetError as e:
                # [Errno 54] Connection reset by peer
                logger.error(f'ConnectionResetError when request_with_unique_key: {e}')
                self.reconnect()
            except ConnectionAbortedError as e:
                logger.error(f'ConnectionAbortedError when request_with_unique_key: {e}')
            except ShouldReconnectException as e:
                logger.error(e)
                self.reconnect()
            except Exception as e:
                logger.error(f'{e}, {traceback.print_exc()}')
            finally:
                pass

    def reconnect(self):
        self._client.reconnect()
        self._checksum = 0
        self._receive_queue.queue.clear()
        self._client.bind()


if __name__ == '__main__':
    host = '127.0.0.1'
    port = 22222
    authkey = b'aaa'
    for _ in range(15):
        manager = RegisterManager(host, port, authkey)
        client = SlaveClient(manager)
        # client.bind()
        worker = SlaveService(client, 3000)
        thread = threading.Thread(target=worker.run, daemon=True)
        thread.start()



import time
import threading
import traceback
from datetime import datetime
from datetime import timezone
from loguru import logger


class ScheduleDictMaker:

    @staticmethod
    def make_from_orms_by_id(orms):
        assert type(orms) == list
        return {orm.id: ScheduleDictMaker.make_from_orm(orm) for orm in orms}

    @staticmethod
    def make_from_orm(orm):
        # print(orm.id)
        _dict = {}
        _dict['id'] = orm.id
        _dict['key'] = orm.key
        _dict['name'] = orm.name
        _dict['sharding_key'] = orm.sharding_key
        _dict['region'] = orm.region
        _dict['stop_time'] = orm.stop_time
        _dict['version'] = orm.version
        _dict['created_time'] = orm.created_time.isoformat()
        _dict['updated_time'] = orm.updated_time.isoformat()
        return _dict


class ScheduleLoader(threading.Thread):

    # global_last_offset是当前的增量加载时间戳

    def __init__(self, repository_read, schedule_queue, load_limit_num=11111):
        threading.Thread.__init__(self)
        self._last_offset = datetime.fromtimestamp(0, timezone.utc)
        self._repository_read = repository_read
        self._schedule_queue = schedule_queue
        self._load_limit_num = load_limit_num
        self._sort_field_name = 'updated_time'
        self._is_active = True
        self.daemon = True

    def log_timer(self, interval):
        try:
            logger.debug(f'ScheduleLoader log, current schedule_queue size: {self._schedule_queue.qsize()}')
        except Exception:
            logger.error(traceback.print_exc())
        finally:
            t = threading.Timer(interval, self.log_timer, args=(interval, ))
            t.daemon = True
            t.start()

    def load_need_run_once(self):
        try:
            need_run_schedule_orms, self._last_offset = self._repository_read.load_schedules_incr(
                self._sort_field_name,
                last_offset=self._last_offset,
                limit=self._load_limit_num,
            )
            for _schedule_orm in need_run_schedule_orms:
                schedule = ScheduleDictMaker.make_from_orm(_schedule_orm)
                schedule_key = schedule['key']
                self._schedule_queue.put(schedule_key)
                logger.debug(f'load_need_run_once put schedule, schedule={schedule}')
            return len(need_run_schedule_orms)
        except Exception as e:
            logger.error(f'load_need_run_once exception {e}, {traceback.print_exc()}')

    def run(self):
        self.log_timer(10)
        while self._is_active:
            try:
                need_run_num = self.load_need_run_once()
                logger.debug(f'need_run_num = {need_run_num}')
            except Exception as e:
                logger.error(f'{e}, {traceback.print_exc()}')
            finally:
                if need_run_num < self._load_limit_num:
                    time.sleep(2)



import sys
import threading
from multiprocessing.managers import BaseManager
from ..common.schedule_center import ScheduleCenter
from ..master.master_manager import MasterManager
from loguru import logger
from ..master.master_api import app
from ..common.protocol import DEFAULT_SCHEDULE_CENTER_NAME
from .source.schedule_repository import ScheduleRepositoryFactory
from .source.schedule_loader import ScheduleLoader


class ScheduleManager(BaseManager):
    pass


class MasterServer():

    def __init__(self, api_port, queue_port, auth_key, db_config, slave_keys_limit=10000, log_level='INFO'):
        self._api_port = api_port
        self._queue_port = queue_port
        self._auth_key = auth_key
        self._db_config = db_config
        logger.remove()
        logger.add(sys.stderr, level=log_level)

    def run(self, schedule_queue, heartbeats_timeout=190, slave_keys_limit=10000):
        # 启动Schedule Center
        schedule_center = ScheduleCenter()
        schedule_center.set_slave_keys_limit(slave_keys_limit)
        mst = MasterManager(schedule_center, schedule_queue, heartbeats_timeout)
        mst.start()
        # 启动Schedule Loader
        repository_read = ScheduleRepositoryFactory.make(self._db_config, 'mysql', None)
        slt = ScheduleLoader(repository_read, schedule_queue, 2)
        slt.start()
        threading.Thread(target=lambda: app.run(host='0.0.0.0', port=self._api_port, debug=True, use_reloader=False), daemon=True).start()
        ScheduleManager.register(DEFAULT_SCHEDULE_CENTER_NAME, ScheduleCenter)
        m = ScheduleManager(address=('0.0.0.0', self._queue_port), authkey=self._auth_key)
        s = m.get_server()
        s.serve_forever()


from sqlalchemy import engine_from_config
from sqlalchemy.orm.session import sessionmaker
from sqlalchemy_utils import database_exists, create_database
from sqlalchemy import event


def before_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    if 'update ' in statement.lower():
        # print('update')
        pass
    if 'delete ' in statement.lower():
        # print('warning delete statement: {}'.format(statement))
        pass
    if 'insert ' in statement.lower():
        pass
    if 'select ' in statement.lower() and 'updated_time' in statement.lower():
        # print(statement)
        pass
    # print('before_cursor_execute')
    # print(statement)


class DbFactory():

    @staticmethod
    def create_engine_from_config(config: dict):
        engine = engine_from_config(config['DB_ENGINE_CONFIG'], prefix=config['DB_ENGINE_CONFIG_PREFIX'])
        if not database_exists(engine.url):
            create_database(engine.url)
        return engine

    @staticmethod
    def create_session_from_config(config: dict):
        engine = DbFactory.create_engine_from_config(config)
        event.listen(engine, 'before_cursor_execute', before_cursor_execute)
        session = sessionmaker()
        session.configure(bind=engine, autoflush=config['DB_SESSION_AUTOFLUSH'], autocommit=config['DB_SESSION_AUTOCOMMIT'])
        s = session()
        return s

    @staticmethod
    def create_engine_and_session_from_config(config: dict):
        engine = DbFactory.create_engine_from_config(config)
        event.listen(engine, 'before_cursor_execute', before_cursor_execute)
        session = sessionmaker()
        session.configure(bind=engine, autoflush=config['DB_SESSION_AUTOFLUSH'], autocommit=config['DB_SESSION_AUTOCOMMIT'])
        s = session()
        return engine, s


class BaseDbManager():

    _engine = None
    _session = None

    def __init__(self, base, config: dict):
        self._base = base
        self._engine, self._session = DbFactory.create_engine_and_session_from_config(config)

    def make_repository(self, db_type, sharding_key):
        raise Exception('BaseDbManager.make_repository should be implement')

    def create_tables(self):
        self._base.metadata.create_all(bind=self._engine)

    def drop_tables(self):
        self._base.metadata.drop_all(bind=self._engine)

    def list_sorted_tables(self):
        return self._base.metadata.sorted_tables

    def get_engine(self):
        return self._engine

    def get_session(self):
        return self._session

    def add(self, model):
        return self._session.add(model)

    def bulk_save_objects(self, models):
        # raise Exception('该方法可能会引起Foraign key的约束造成插入失败')
        return self._session.bulk_save_objects(models)

    def commit(self):
        return self._session.commit()

    def close(self):
        self._session.close()

    def dispose(self):
        self._engine.dispose()



import time
import traceback
import socket
from ..common.tools import get_kpi
from loguru import logger


class SlaveClient():

    def __init__(self, register_manager):
        """
        SlaveClient是对ScheduleCenter可用方法的封装，register_manager相当于schedule_center的代理
        """
        self._register_manager = register_manager
        self._address = None
        self._secret_key = None

    def bind(self):
        hostname = socket.gethostname()
        ip_address = socket.gethostbyname(hostname)
        address_suffix = str(int(time.time() * 1000))[-9:]
        self._address = f'{hostname}-{ip_address}-{address_suffix}'
        print(f'bind address: {self._address}')
        _current_kpi = get_kpi()
        code, result = self._register_manager.get_schedule_center().bind(self._address, _current_kpi)
        if code != 200:
            raise Exception(f'exception when bind schedule center, code = {code}, result = {result}')
        self._secret_key = result

    def take(self, checksum, batch_num=2):
        _current_kpi = get_kpi()
        logger.debug(f'kpi = {_current_kpi}')
        result = self._register_manager.get_schedule_center().take(self._address, self._secret_key, checksum, _current_kpi, batch_num=batch_num, now=None)
        return result

    def reconnect(self):
        self._register_manager.reconnect()



import time
import traceback
import threading
from ..common import tools
from loguru import logger
from ..common.schedule_center import ScheduleCenter


class MasterManager(threading.Thread):
    """
    MasterManager负责对ScheduleCenter单例进行治理
    Master以线程形式运行。
    Master作用：
    1 监控Master的绑定Worker是否正常接收任务和心跳，如果不正常，则重新分配；
    2 监控Master是否有需要重新分配的任务，有的话则重新分配；
    3 监控Master的mapping关系是否与Worker的mapping关系一致，如果不一致，告警（理论上不应该长时间存在这种情况）；
    """

    def __init__(self, schedule_center: ScheduleCenter, schedule_queue, heartbeats_timeout, batch_num=10000):
        threading.Thread.__init__(self)
        self._schedule_center = schedule_center
        self._schedule_queue = schedule_queue
        self._heartbeats_timeout = heartbeats_timeout
        self._batch_num = batch_num
        self._is_active = True
        self.daemon = True

    def run_once(self, now=None):
        timeout_address_set = self._step1_1_get_timeout_slave_address_set(now)
        self._step1_2_unbind_timeout_slave_address_set(timeout_address_set)
        added_key_num = self._step2_1_add_need_schedule_key_set()
        self._step2_2_reallocate()
        return added_key_num

    def _step1_1_get_timeout_slave_address_set(self, now=None):
        if now is None:
            now = int(time.time())
        timeout_address_set = set()
        for _address, _timestamps, in self._schedule_center.slave_heartbeats().items():
            if now - _timestamps > self._heartbeats_timeout:
                timeout_address_set.add(_address)
        return timeout_address_set

    def _step1_2_unbind_timeout_slave_address_set(self, slave_address_set):
        for _address in slave_address_set:
            self.timeout_unbind(_address)
            logger.warning(f'timeout_unbind address: {_address}')

    def _step2_1_add_need_schedule_key_set(self):
        key_list = tools.getn_from_queue(self._schedule_queue, self._batch_num)
        if len(key_list) > 0:
            key_set = set(key_list)
            self._schedule_center.add_reallocate_keys(key_set)
            return len(key_list)
        return 0

    def _step2_2_reallocate(self):
        for key in self._schedule_center.reallocate_keys().copy():
            result = self._schedule_center.allocate(key)
            if result is True:
                self._schedule_center.reallocate_keys().remove(key)
                logger.debug(f'reallocate key: {key}, result = {result}')
                # logger.info(f'reallocate key: {key}, result = {result}')
            else:
                logger.error(f'reallocate key: {key}, result = {result}')
                time.sleep(1)

    def timeout_unbind(self, address):
        secret_key_store = self._schedule_center.secret_key_store()
        secret_key = secret_key_store[address]
        self._schedule_center.unbind(address, secret_key)

    def run(self):
        assert self._is_active is True
        while self._is_active:
            try:
                num = self.run_once(now=None)
                if num == 0:
                    time.sleep(1)
            except Exception as e:
                logger.error(f'{e}, {traceback.print_exc()}')
                time.sleep(3)
            finally:
                pass



import time
from multiprocessing.managers import BaseManager
from multiprocessing.managers import BaseProxy
from ..common.protocol import DEFAULT_SCHEDULE_CENTER_NAME
from loguru import logger


class RegisterManager:

    def __init__(self, host, port, authkey):
        logger.debug(f'RegisterManager init host={host}, port={port}')
        BaseManager.register(DEFAULT_SCHEDULE_CENTER_NAME)
        self._address = (host, port)
        self._auth_key = authkey
        self._m = BaseManager(address=self._address, authkey=self._auth_key)
        self.rebind()

    def get_schedule_center(self):
        return self._schedule_center

    def rebind(self):
        # 由于reconnect的时候也会用到，所以单独抽出来了
        self._m.connect()
        self._schedule_center = getattr(self._m, DEFAULT_SCHEDULE_CENTER_NAME)()

    def reconnect(self, max_retry_times=10):
        for times in range(max_retry_times):
            try:
                if self._address in BaseProxy._address_to_local:
                    if getattr(BaseProxy._address_to_local[self._address][0], 'connection', False):
                        del BaseProxy._address_to_local[self._address][0].connection
                        logger.warning('Delete connection success')
                    self.rebind()
                    logger.warning(f'RegisterManager reconnect success, times = {times}')
                    return True
            except Exception as e:
                retry_after = 1 * times
                logger.error(f'reconnect error: {e}, times = {times}, retry after {retry_after} seconds')
                time.sleep(retry_after)
        return False